from pasteob import *
from pasteob.auth.logic import *
from hashlib import md5
import time, random

__all__ = ['DigestAuthLogic']



_nonces = {}


def calc_ha1(realm, username, password):
    return md5("%s:%s:%s" % (username, realm, password)).hexdigest()

##    If the "algorithm" directive's value is "MD5-sess", then A1 is
##    calculated only once - on the first request by the client following
##    receipt of a WWW-Authenticate challenge from the server.  It uses the
##    server nonce from that challenge, and the first client nonce value to
##    construct A1 as follows:
##
##       A1       = H( unq(username-value) ":" unq(realm-value)
##                      ":" passwd )
##                      ":" unq(nonce-value) ":" unq(cnonce-value)
##

##    If the "qop" value is "auth-int", then A2 is:
##
##       A2       = Method ":" digest-uri-value ":" H(entity-body)
##


##         AuthenticationInfo = "Authentication-Info" ":" auth-info
##         auth-info          = 1#(nextnonce | [ message-qop ]
##                                | [ response-auth ] | [ cnonce ]
##                                | [nonce-count] )
##         nextnonce          = "nextnonce" "=" nonce-value
##         response-auth      = "rspauth" "=" response-digest
##         response-digest    = <"> *LHEX <">
##

def random_str():
    return md5("%s:%s" % (time.time(), random.random())).hexdigest()

class DigestAuthLogic(AuthLogic):
    # if populate_req is True the environ keys 'REMOTE_USER' and 'AUTH_TYPE' will be
    # adjusted if the provided credentials are valid
    populate_req = True
    realm = None
    _username = None # this is what user claims -- can be invalid!
    _realm = None
    def __init__(self, req):
        assert self.realm, "DigestAuthLogic subclasses should define their realm"
        self.stale = False
        self.method = req.method
        self.path_qs = req.path_qs
        if req.authorization and req.authorization[0] == 'Digest':
            auth = self._auth = req.authorization[1]
            self._username = auth.get('username')
            self._realm = auth.get('realm')
            if self._username and self._realm == self.realm: # username can't be empty
                ha1 = self.get_ha1()
                if ha1:
                    try:
                        self.check_auth(ha1, **self._auth)
                    except TypeError: # missing keys in auth dict
                        raise HTTPBadRequest("Missing keys in auth dict: %s" % auth)
            if self.logged_in and self.populate_req:
                req.remote_user = self._username
                #req.environ['AUTH_TYPE'] = 'digest'

    def get_ha1(self): # HA1 to expect from the request
        password = self.get_password()
        if password:
            return self.calc_ha1(self._username, password)

    def get_password(self): # password for the user specified in the request
        raise NotImplementedError

    def calc_ha1(self, username, password):
        return calc_ha1(self.realm, username, password)

    def check_auth(self, ha1, response, nonce, uri, qop, cnonce, nc, **kw):
        if qop != 'auth':
            raise HTTPBadRequest("Invalid qop=%s" % qop)
        elif self.path_qs != uri:
            raise HTTPBadRequest("Request for %s tries to authenticate as a request to %s"
                % (self.path_qs, uri)
            )
        ha2 = md5('%s:%s' % (self.method, uri)).hexdigest()
        chk = "%s:%s:%s:%s:%s:%s" % (ha1, nonce, nc, cnonce, qop, ha2)
        pnc = _nonces.pop(nonce, '00000000')
        if response != md5(chk).hexdigest():
            pass # bad password
        elif nc <= pnc:
            self.stale = True # might be a replay or brute-force attack
        else:
            _nonces[nonce] = nc
            self.logged_in = True

    def _make_user(self):
        return User(self._username)

    _generate_nonce = staticmethod(random_str)

    def _login_now(self):
        assert not self.logged_in
        nonce = self._generate_nonce()
        _nonces[nonce] = None #@@?
        challenge = dict(
            #domain = <"> URI ( 1*SP URI ) <"> (URI = absoluteURI | abs_path)
            #algorithm ( "MD5" | "MD5-sess" | token )
            qop='auth', #"auth" | "auth-int" | token
            realm=self.realm,
            nonce=nonce,
            opaque=random_str(), #@@ check that client sends this
            stale=('true' if self.stale else 'false')
        )
        return HTTPUnauthorized(www_authenticate=('Digest', challenge))



