from webob.exc import HTTPTemporaryRedirect, HTTPSeeOther
from pasteob.reqaddon import ReqAddon
from pasteob.creq import creq, creq_mw
from mext.context import _ClassDelegate, Setting

__all__ = [
    'WebAuth', 'AuthLogic', 'User',
    'creq_mw',
]

webauth_logic = Setting()

class WebAuth(_ClassDelegate):
    """
        Every auth logic should be able to tell from request
        if the user is logged in and if so, produce a correspoding
        User object.

        Regarding login there are three cases:

         a) The logic is not capable of logging user in. For example
            cookie-based auth logic can validate the cookie, but knows
            nothing about asking the user for credentials or checking them.

         b) The logic can override any response with a request for credentials.
            For example Basic and Digest authentication can return 403 Not Authorized
            responses at any URL and expect the next request to have the auth data
            they can validate.
            Subclasses of such logic normally provide `check_password()` method
            that checks ._username and ._password attrs that are set to credentials
            provided by the user. It then should set self.logged_in and self._user
            if the credentials are valid.

         c) The logic might only be able to request for credentials and authenticate
            the user at some specific URLs. In fact it might not contain the login
            logic, but defer it to some other app. What it provides is an URL the
            user must be directed to if he is to be authenticated.
            After authentication the user would be returned to the original URL
            (but that can be overridden).

            For example, think of an auth system that only allows login via HTTPS
            or a login process that needs multiple steps. This is the most common case.

            If the logic provides this functionality, it is automatically capable of (b)
            The overriding response in this case is a temporary redirect to a login URL.

        Also, in practice logics of type-a are augmented in a subclass to provide a login URL
        (making them also type-c) and mount a login app at that URL that authenticates the user by
        means of some other logic of type-b or in some other way and then sets the cookies
        that the base type-a logic can understand.

    """
    # properties and methods to be implemented in every logic
    logged_in = user = None
    login_url = logout_url = None
    require_user = None

    def __new__(cls, req):
        logic_factory = webauth_logic.get()
        return logic_factory(req)

    @staticmethod
    def activate(logic):
        webauth_logic.set(logic)

    @staticmethod
    def get():
        # identical to return cls(creq.get())
        req = creq.get()
        logic_factory = webauth_logic.get()
        return logic_factory(req)



def not_implemented(*args):
    raise NotImplementedError

class AuthLogic(ReqAddon):
    logged_in = False
    _user = None
    _login_url = _logout_url = _make_user = not_implemented
    _login_now = _logout_now = None
    def __new__(cls, req):
        # this allows us not to require the subclasses to call superclass' __init__
        r = object.__new__(cls)
        r.url = req.url
        return r

    def login_url(self, return_to=None):
        assert not self.logged_in
        return self._login_url(return_to or self.url)

    def logout_url(self, return_to=None):
        assert self.logged_in
        return self._logout_url(return_to or self.url)

    @property
    def user(self):
        if not self.logged_in:
            return None
        if self._user is None:
            self._user = self._make_user()
        return self._user

    def require_user(self, user_id=None):
        """
            Return a User object or override response with a login page.
        """
        if self.logged_in:
            if user_id is None or self.user.user_id == user_id:
                return self.user
        self.login_now()

    def login_now(self):
        assert not self.logged_in
        if self._login_now:
            raise self._login_now()
        else:
            raise HTTPSeeOther(location=self.login_url())
            #raise HTTPTemporaryRedirect(location=self.login_url())



class IUser(object):
    user_id = None # opaque unique identifier
    nickname = None # a name to address the user in UI. not unique
    email = None # verified user email (or None). can change for a given user_id
    roles = None # frozenset of strings


class User(object):
    def __init__(self, user_id, email=None, nickname=None, roles=()):
        self.user_id = user_id
        self.email = email
        if not nickname:
            nickname = email.split('@')[0] if email else user_id
        self.nickname = nickname
        self.roles = frozenset(roles)

