"""
Middleware that displays everything that is printed inline in
application pages.

Anything printed during the request will get captured and included on
the page.  It will usually be included as a floating element in the
top right hand corner of the page.  If you want to override this
you can include a tag in your template where it will be placed::

  <pre id="paste-debug-prints"></pre>

You might want to include ``style="white-space: normal"``, as all the
whitespace will be quoted, and this allows the text to wrap if
necessary.
"""

from __future__ import with_statement
import sys, os, cgi, re
from cStringIO import StringIO
from contextlib import contextmanager
from mext.threads import RLock
from pasteob import *

__all__ = ['CatchPrints']

_lock = RLock()

# are we running on Google App Engine?
_gae_deployed = os.environ.get('SERVER_SOFTWARE', '').startswith('Google')

class CatchPrints(object):
    def __init__(self, app, replace_stdout=False, print_wsgi_errors=False, use_lock=True):
        self.app = app
        if _gae_deployed:
            replace_stdout = True
            use_lock = False
            #raise RuntimeError("CatchPrints is not compatible with GAE at the moment")
        self.replace_stdout = replace_stdout
        self.print_wsgi_errors = print_wsgi_errors
        self.use_lock = use_lock

    @webob_wrap
    def __call__(self, req):
        with self.new_stdout(req) as f:
            r = req.get_response(self.app)
            r.body # exhausts app_iter
        if (r.content_type or '').startswith('text/html'):
            self.add_log(r, f.getvalue())
        elif self.replace_stdout and not self.print_wsgi_errors:
            # Then the prints will be lost, unless...
            req.environ['wsgi.errors'].write(f.getvalue())
        return r

    @contextmanager
    def new_stdout(self, req):
        # we use the lock to avoid concurrent
        # apps from replacing each others' stdout
        lock = _lock if self.use_lock else _dummy_lock()
        with lock:
            stdout = sys.stdout
            # file obj to track all stdout output for this req
            f = StringIO()
            mf = MultiFile([f])
            if not self.replace_stdout:
                mf.add(stdout)
            if self.print_wsgi_errors:
                mf.add(req.environ['wsgi.errors'])

            try:
                sys.stdout = mf
                yield f # return only the req-specific file, not real stdout
            finally:
                sys.stdout = stdout


    _rx_body = re.compile(r'<body[^>]*>', re.I)
    _rx_explicit = re.compile(r'<pre\s*[^>]*id="paste-debug-prints".*?>', re.I+re.S)

    log_template = (
        '<pre style="padding:1ex; border: 1px solid #000; white-space: normal; '
        'background-color: #ffd; color: #000; float: right; font-size: 9pt">'
        '<b style="border-bottom: 1px solid #000">Log messages</b><br>'
        '%s</pre>'
    )

    def add_log(self, resp, log):
        if log:
            html = resp.body
            text = cgi.escape(log)
            text = text.replace('\n', '<br>')
            text = text.replace('  ', '&nbsp; ')
            match = self._rx_explicit.search(html)
            if not match:
                text = self.log_template % text
                match = self._rx_body.search(html)
            if match:
                resp.body = html[:match.end()] + text + html[match.end():]
            else:
                resp.body = text + html

@contextmanager
def _dummy_lock():
    yield



class MultiFile(object):
    def __init__(self, files=()):
        self.files = list(files)

    def add(self, file):
        self.files.append(file)

    def write(self, v):
        if isinstance(v, unicode):
            v = str(v) # WSGI is picky in this case
        for file in self.files:
            file.write(v)







